# Apply a watermark to a PDF file
import os
import shutil
from tempfile import TemporaryDirectory
from looptools import Timer
from pdf.utils import add_suffix, open_window, Receipt, Info
from pdf.modify.draw import WatermarkDraw
from pdf.modify.canvas import CanvasConstructor
from pdf.conduit.lib import IMAGE_DEFAULT, IMAGE_DIRECTORY
from pdf.conduit.encrypt import Encrypt
from pdf.conduit.watermark.add import WatermarkAdd


class Watermark:
    def __init__(self, document, remove_temps=True, move_temps=None, open_file=False, tempdir=None, receipt=None,
                 use_receipt=True, progress_bar_enabled=False, progress_bar='tqdm'):
        """
        Watermark and encrypt a PDF document.

        Manage watermarking processes from single class initialization.  This class utilizes the draw,
        add and encrypt modules.

        :param document: str
            PDF document full path
        :param remove_temps: bool
            Remove temporary files after completion
        :param open_file: bool
            Open file after completion
        :param tempdir: function or str
            Temporary directory for file writing
        :param receipt: cls
            Use existing Receipt object if already initiated
        :param use_receipt: bool
            Print receipt information to console and write to file
        """
        self.time = Timer()
        self.document_og = document
        self.document = self.document_og
        self.watermark = None
        self.remove_temps = remove_temps
        self.move_temps = move_temps
        self.open_file = open_file

        if not tempdir:
            self._temp = TemporaryDirectory()
            self.tempdir = self._temp.name
        elif isinstance(tempdir, TemporaryDirectory):
            self._temp = tempdir
            self.tempdir = self._temp.name
        else:
            self.tempdir = tempdir

        self.progress_bar_enabled = progress_bar_enabled
        self.progress_bar = progress_bar

        self.use_receipt = use_receipt
        if use_receipt:
            if isinstance(receipt, Receipt):
                self.receipt = receipt
            else:
                self.receipt = Receipt(use_receipt).set_dst(document)

    def __str__(self):
        return str(self.document)

    def cleanup(self):
        runtime = self.time.end
        if self.use_receipt:
            self.receipt.add('~run time~', runtime)
            self.receipt.dump()
        if self.move_temps:
            if os.path.isdir(self.move_temps):
                shutil.move(self.tempdir, self.move_temps)
        if self.remove_temps:
            if os.path.isdir(self.tempdir):
                shutil.rmtree(self.tempdir)
        else:
            open_window(self.tempdir)
        return self.document

    def draw(self, text1=None, text2=None, copyright=True, image=IMAGE_DEFAULT, rotate=30, opacity=0.08, compress=0,
             flatten=False, add=False):
        """
        Draw watermark PDF file.

        Create watermark using either a reportlabs canvas or a PIL image.

        :param text1: str
            Text line 1
        :param text2: str
            Text line 2
        :param copyright: bool
            Draw copyright and year to canvas
        :param image: str
            Logo image to be used as base watermark
        :param rotate: int
            Degrees to rotate canvas by
        :param opacity: float
            Watermark opacity
        :param compress: bool
            Compress watermark contents  (not entire PDF)
        :param flatten: bool
            Draw watermark with multiple layers or a single flattened layer
        :param add: bool
            Add watermark to original document
        :return: str
            Watermark PDF file full path
        """
        im_path = os.path.join(IMAGE_DIRECTORY, image)
        if os.path.isfile(im_path):
            image = im_path

        # Add to receipt
        if self.use_receipt:
            self.receipt.add('Text1', text1)
            self.receipt.add('Text2', text2)
            self.receipt.add('Image', os.path.basename(image))
            self.receipt.add('WM Opacity', str(int(opacity * 100)) + '%')
            self.receipt.add('WM Compression', compress)
            self.receipt.add('WM Flattening', flatten)

        co = CanvasConstructor(text1, text2, copyright, image, rotate, opacity, tempdir=self.tempdir)
        objects, rotate = co.img() if flatten else co.canvas()  # Run img constructor method if flatten is True

        # Draw watermark to file
        self.watermark = WatermarkDraw(objects, rotate=rotate, compress=compress, tempdir=self.tempdir,
                                       pagesize=Info(self.document_og).size, pagescale=True).write()

        if not add:
            return self.watermark
        else:
            self.add()
            return self.cleanup()

    def add(self, document=None, watermark=None, underneath=False, output=None, suffix='watermarked', method='pdfrw'):
        """
        Add a watermark file to an existing PDF document.

        Rotate and upscale watermark file as needed to fit existing PDF document.  Watermark can be overlayed or
        placed underneath.

        :param document: str
            PDF document full path
        :param watermark: str
            Watermark PDF full path
        :param underneath: bool
            Place watermark either under or over existing PDF document
        :param output: str
            Output file path
        :param suffix: str
            Suffix to append to existing PDF document file name
        :param method: str
            PDF library to be used for watermark adding
        :return: str
            Watermarked PDF Document full path
        """
        if self.use_receipt:
            self.receipt.add('WM Placement', 'Overlay')
        if not watermark:
            watermark = self.watermark
        if not document:
            document = self.document
        self.document = str(WatermarkAdd(document, watermark, output=output, underneath=underneath,
                                         tempdir=self.tempdir, suffix=suffix, method=method))
        if self.use_receipt:
            self.receipt.add('Watermarked PDF', os.path.basename(self.document))
        if self.open_file:
            open_window(self.document)
        return self.document

    def encrypt(self, user_pw='', owner_pw=None, encrypt_128=True, allow_printing=True, allow_commenting=False,
                document=None):
        """
        Encrypt a PDF document to add passwords and restrict permissions.

        Add a user password that must be entered to view document and a owner password that must be entered to alter
        permissions and security settings.  Encryption keys are 128 bit when encrypt_128 is True and 40 bit when
        False.  By default permissions are restricted to print only, when set to false all permissions are allowed.
        TODO: Add additional permission parameters

        :param user_pw: str
            User password required to open and view PDF document
        :param owner_pw: str
            Owner password required to alter security settings and permissions
        :param encrypt_128: bool
            Encrypt PDF document using 128 bit keys
        :param allow_printing: bool
            Restrict permissions to print only
        :return: str
            Encrypted PDF full path
        """
        document = self.document if document is None else document
        if self.use_receipt:
            self.receipt.add('User pw', user_pw)
            self.receipt.add('Owner pw', owner_pw)
            if encrypt_128:
                self.receipt.add('Encryption key size', '128')
            else:
                self.receipt.add('Encryption key size', '40')
            if allow_printing:
                self.receipt.add('Permissions', 'Allow printing')
            else:
                self.receipt.add('Permissions', 'Allow ALL')
        p = str(Encrypt(document, user_pw, owner_pw, output=add_suffix(self.document_og, 'secured'),
                        bit128=encrypt_128, allow_printing=allow_printing, allow_commenting=allow_commenting,
                        progress_bar_enabled=self.progress_bar_enabled, progress_bar=self.progress_bar))
        if self.use_receipt:
            self.receipt.add('Secured PDF', os.path.basename(p))
        return p



__all__ = ["Watermark", "WatermarkAdd", "Label"]


from pdf.conduit.watermark.watermark import Watermark
from pdf.conduit.watermark.add import WatermarkAdd
from pdf.conduit.watermark.label import Label



import os
from PyBundle import bundle_dir, resource_path


def _image_directory():
    directory = os.path.join(bundle_dir(), 'img')
    if os.path.exists(directory):
        return directory
    else:
        print(directory, 'can not be found')


IMAGE_DIRECTORY = _image_directory()


def available_images():
    imgs = [i for i in os.listdir(IMAGE_DIRECTORY) if not i.startswith('.')]
    if len(imgs) > 0:
        return sorted(imgs, reverse=True)
    else:
        return ['Add images...']


IMAGE_DEFAULT = resource_path('Wide.png')


__all__ = ['IMAGE_DEFAULT', 'IMAGE_DIRECTORY', 'available_images']


# Add a watermark PDF file to another PDF file
from tempfile import NamedTemporaryFile
from PyBundle import resource_path
from pdfrw import PdfReader, PdfWriter, PageMerge
from reportlab.lib.pagesizes import letter
from PyPDF3 import PdfFileReader, PdfFileWriter
from PyPDF3.pdf import PageObject
from pdf.transform.upscale import upscale
from pdf.transform.rotate import rotate
from pdf.utils import add_suffix, Info


class WatermarkAdd:
    def __init__(self, document, watermark, underneath=False, overwrite=False, output=None, suffix='watermarked',
                 decrypt=False, tempdir=None, method='pdfrw'):
        """
        Add a watermark to an existing PDF document

        1. Read PDF document and decrypt if encrypted
        2. Get input PDF document info
            2a. Retrieve dimensions
            2b. Determine orientation
            2c. Upscale to fit letter
        3. Get watermark info
            3a. Retrieve dimensions
            3b. Rotate if necessary
            3c. Upscale to fit document
        4. Set file names
            4a. Check if a upscaled doc was created, if not use original document
            4b. Check if watermark was rotated, if not use original watermark
        5. Add watermark to PDF document
            5a. Create watermarked document filename (ex: originalpdfname_watermarked.pdf)
            5b. Create PDF file reader and file writer objects
            5c. Add watermark to each page of original
            5d. Save watermarked document to file
        """
        self.rotate = 0
        self.scale = 0
        self.underneath = underneath
        self.tempdir = tempdir
        self.method = method

        self.document_reader = self._document_reader(document, decrypt)
        self.document = self._get_document_info(document)
        self.watermark_file = self._get_watermark_info(self.document, watermark)
        pdf_fname, wtrmrk_fname = self._set_filenames

        if overwrite:
            self.output_filename = document
        elif output:
            self.output_filename = output
        elif suffix:
            self.output_filename = add_suffix(document, suffix)
        else:
            tmpf = NamedTemporaryFile(suffix='.pdf', dir=self.tempdir, delete=False)
            self.output_filename = resource_path(tmpf.name)

        self.add(pdf_fname, wtrmrk_fname)

    def __str__(self):
        return str(self.output_filename)

    @staticmethod
    def _document_reader(document, decrypt=False):
        # 1. Read PDF document and create file reader object
        if decrypt:
            reader = PdfFileReader(document)
            reader.decrypt(decrypt)
            return reader
        else:
            return PdfFileReader(document)

    def _get_document_info(self, filename):
        pdf_file = {'path': filename}

        # 2a. Get PDF width and height
        pdf_file.update(Info(self.document_reader).dimensions)

        # 2b. Get PDF file orientation
        if pdf_file['h'] > pdf_file['w']:
            pdf_file['orientation'] = 'portrait'
            letter_size = {'w': int(letter[0]), 'h': int(letter[1])}
        else:
            pdf_file['orientation'] = 'landscape'
            letter_size = {'h': int(letter[0]), 'w': int(letter[1])}

        # 2c. Upscale PDF if it is smaller than a letter
        if pdf_file['w'] <= letter_size['w'] or pdf_file['h'] <= letter_size['h']:
            scale = float(letter_size['w'] / pdf_file['w'])
            pdf_file['upscaled'] = upscale(pdf_file['path'], scale=scale, tempdir=self.tempdir)
            self.document_reader = self._document_reader(pdf_file['upscaled'])
        return pdf_file

    def _get_watermark_info(self, document, watermark, margin_x=0, margin_y=0):
        # 3a. Get watermark path and dimensions
        watermark_file = {'path': watermark}
        watermark_file.update(Info(watermark).dimensions)

        # 3b. Check if watermark file needs to be rotated
        if watermark_file['w'] > watermark_file['h'] and document['orientation'] is 'portrait':
            self.rotate = 90
            watermark_file['rotated'] = rotate(watermark, self.rotate, tempdir=self.tempdir, method=self.method)

        # Set watermark file to be used for upscaling
        try:
            wtrmrk = watermark_file['rotated']
            watermark_file.update(Info(watermark_file['rotated']).dimensions)
        except KeyError:
            wtrmrk = watermark_file['path']

        # 3c. Check if watermark file needs to be upscaled
        scale = 0
        if watermark_file['w'] <= document['w']:
            scale = float(document['w'] / watermark_file['w'])

            if watermark_file['h'] * scale > document['h']:
                scale = float(document['h'] / watermark_file['h'])

            watermark_file['upscaled'] = upscale(wtrmrk, scale=scale, tempdir=self.tempdir, method=self.method)
        self.scale = scale
        return watermark_file

    @property
    def _set_filenames(self):
        # 4a. If upscaled PDF file does not exists use input PDF path
        try:
            pdf = self.document['upscaled']
        except KeyError:
            pdf = self.document['path']

        # 4b. If rotated watermark file does not exists use input watermark path
        try:
            try:
                watermark = self.watermark_file['upscaled']
            except KeyError:
                watermark = self.watermark_file['rotated']
        except KeyError:
            watermark = self.watermark_file['path']
        return pdf, watermark

    def add(self, document, watermark):
        """Add watermark to PDF by merging original PDF and watermark file."""
        # 5a. Create output PDF file name
        output_filename = self.output_filename

        def pypdf3():
            """Much slower than PyPDF3 method."""
            # 5b. Get our files ready
            document_reader = PdfFileReader(document)
            output_file = PdfFileWriter()

            # Number of pages in input document
            page_count = document_reader.getNumPages()

            # Watermark objects
            watermark_reader = PdfFileReader(watermark)
            wtrmrk_page = watermark_reader.getPage(0)
            wtrmrk_width = (wtrmrk_page.mediaBox.getWidth() / 2) + 0
            wtrmrk_height = (wtrmrk_page.mediaBox.getHeight() / 2) + 80
            wtrmrk_rotate = -int(Info(watermark_reader).rotate) if Info(watermark_reader).rotate is not None else 0

            # 5c. Go through all the input file pages to add a watermark to them
            for page_number in range(page_count):
                # Merge the watermark with the page
                if not self.underneath:
                    input_page = document_reader.getPage(page_number)
                    if wtrmrk_rotate is not 0:
                        input_page.mergeRotatedTranslatedPage(wtrmrk_page, wtrmrk_rotate, wtrmrk_width, wtrmrk_height)
                    else:
                        wtrmrk_width = 0
                        wtrmrk_height = 0
                        input_page.mergeTranslatedPage(wtrmrk_page, wtrmrk_width, wtrmrk_height)
                else:
                    size = Info(document_reader).dimensions
                    input_page = PageObject().createBlankPage(document_reader, size['w'], size['h'])
                    if wtrmrk_rotate is not 0:
                        input_page.mergeRotatedTranslatedPage(wtrmrk_page, wtrmrk_rotate, wtrmrk_width, wtrmrk_height)
                    else:
                        wtrmrk_width = 0
                        wtrmrk_height = 0
                        input_page.mergeTranslatedPage(wtrmrk_page, wtrmrk_width, wtrmrk_height)
                    input_page.mergePage(document_reader.getPage(page_number))

                # Add page from input file to output document
                output_file.addPage(input_page)

            # 5d. finally, write "output" to PDF
            with open(output_filename, "wb") as outputStream:
                output_file.write(outputStream)
            return output_filename

        def pdfrw():
            """Faster than PyPDF3 method by as much as 15x."""
            # TODO: Fix issue where watermark is improperly placed on large pagesize PDFs
            # print(Info(document).size)
            # print(Info(watermark).size)
            # print('\n')

            # Open both the source files
            wmark_trailer = PdfReader(watermark)
            trailer = PdfReader(document)

            # Handle different sized pages in same document with
            # a memoization cache, so we don't create more watermark
            # objects than we need to (typically only one per document).

            wmark_page = wmark_trailer.pages[0]
            wmark_cache = {}

            # Process every page
            for pagenum, page in enumerate(trailer.pages, 1):

                # Get the media box of the page, and see
                # if we have a matching watermark in the cache
                mbox = tuple(float(x) for x in page.MediaBox)
                odd = pagenum & 1
                key = mbox, odd
                wmark = wmark_cache.get(key)
                if wmark is None:

                    # Create and cache a new watermark object.
                    wmark = wmark_cache[key] = PageMerge().add(wmark_page)[0]

                    # The math is more complete than it probably needs to be,
                    # because the origin of all pages is almost always (0, 0).
                    # Nonetheless, we illustrate all the values and their names.

                    page_x, page_y, page_x1, page_y1 = mbox
                    page_w = page_x1 - page_x
                    page_h = page_y1 - page_y  # For illustration, not used

                    # Scale the watermark if it is too wide for the page
                    # (Could do the same for height instead if needed)
                    if wmark.w > page_w:
                        wmark.scale(1.0 * page_w / wmark.w)

                    # Always put watermark at the top of the page
                    # (but see horizontal positioning for other ideas)
                    wmark.y += page_y1 - wmark.h

                    # For odd pages, put it at the left of the page,
                    # and for even pages, put it on the right of the page.
                    if odd:
                        wmark.x = page_x
                    else:
                        wmark.x += page_x1 - wmark.w

                    # Optimize the case where the watermark is same width
                    # as page.
                    if page_w == wmark.w:
                        wmark_cache[mbox, not odd] = wmark

                # Add the watermark to the page
                PageMerge(page).add(wmark, prepend=self.underneath).render()

            # Write out the destination file
            PdfWriter(output_filename, trailer=trailer).write()

        if self.method is 'pypdf3':
            return pypdf3()
        else:
            return pdfrw()



from pdf.modify.canvas import CanvasObjects, CanvasStr
from pdf.modify.draw import WatermarkDraw
from pdf.utils import Info, add_suffix
from pdf.conduit.watermark import Watermark


def mean(numbers):
    return float(sum(numbers)) / max(len(numbers), 1)


class Label(WatermarkDraw):
    def __init__(self, document, label, title_page=False, suffix='labeled', output=None, tempdir=None):
        self.size = Info(document).size
        super(Label, self).__init__(self._create_canvas_objects(label, title_page),
                                    pagesize=self.size, tempdir=tempdir)
        self.document = document
        self.watermark = self._write()

        if output:
            self.output = output
        elif suffix:
            suffix = label if not suffix else suffix
            self.output = add_suffix(self.document, suffix)
        else:
            self.output = self.dst

    def _create_canvas_objects(self, label, title_page):
        objects = CanvasObjects()
        if not title_page:
            objects.add(CanvasStr(label, size=int(mean(self.size) * .02), opacity=1,
                                  x=-(self.size[0] / 2) + 15,
                                  y=-(self.size[1] / 2) + 25, x_centered=False))
        else:
            objects.add(CanvasStr(label, size=int(mean(self.size) * .1), opacity=1, y_centered=True))
        return objects

    def write(self, cleanup=True):
        wm = Watermark(self.document, tempdir=self.dir, use_receipt=False, open_file=False, remove_temps=True)
        labeled = wm.add(watermark=self.watermark, output=self.output)
        if cleanup:
            wm.cleanup()
        return labeled


__all__ = []


# Conduit installation
try:
    from pdf.conduit import Encrypt, Watermark, Label, WatermarkAdd
    CONDUIT_INSTALL = True
    __all__.extend(["Encrypt", "Watermark", "Label", "WatermarkAdd"])
except ImportError:
    CONDUIT_INSTALL = False


# Utils installation
try:
    from pdf.utils.info import Info
    UTILS_INSTALL = True
    __all__.extend(["Info"])
except ImportError:
    UTILS_INSTALL = False


# GUI installation
try:
    from pdf.gui.gui import GUI
    GUI_INSTALLED = True
    __all__.extend(["GUI"])
except ImportError:
    GUI_INSTALLED = False


# Transform installation
try:
    from pdf.transform import upscale, rotate, slicer, Merge
    MODIFY_INSTALLED = True
    __all__.extend(["slicer", "upscale", "rotate", "Merge"])
except ImportError:
    MODIFY_INSTALLED = False


# Convert installation
try:
    from pdf.convert import Flatten
    CONVERT_INSTALLED = True
    __all__.extend(["Flatten"])
except ImportError:
    CONVERT_INSTALLED = False
